/*
                       cube.c
                       ======
               Uwe Berger; 2007, 2008
                 
Aufgaben:
---------
-Hardwareansteuerung des 3D-Cube via Interrupt-Routine, welche Ebenen, Spalten
 und Zeilen hochzaehlt sowie Inhalte an Cube-Hardware ausgibt
-Kommunikationsprotokoll (via Ethernet), um hauptsaechlich die Bildinhalte des
 3D-Cube an Etherrrape zu uebertragen

Hardware: 
---------
Decoder mit 3 hintereinander geschalteten Schieberegistern. 
Portbelegung siehe cube.h

*/
#include <string.h>
#include "cube.h"
#include <avr/io.h>
#include <avr/interrupt.h>


/* Timer */
volatile uint16_t timer_preload  = T1_PRELOAD;

/* interne Variable */
volatile uint8_t spalte    = 0;
volatile uint8_t ebene     = 0;
volatile uint8_t led[CUBE_XYZ];


/* Interruptroutine Steuerung Cube-Hardware */
ISR(TIMER1_OVF_vect)
{
	/* Preload-Wert des Timers neu setzen */
	TCNT1   =  timer_preload;
	
	/*************************************************************************/
	/* vorhergehende Zeile sauber auf LOW schalten, wegen "LED-Nachleuchten" */
	/*************************************************************************/
		/* dazu Schieberegisterkette loeschen (Low-Impulse auf SCL) */
	PORT_SR &= ~(1 << PIN_SCL);	/* SCL auf 0 */
	PORT_SR |=  (1 << PIN_SCL);	/* SCL auf 1 */
		/* 0-1-Flanke an RCK bringt die Daten aus den Registern in die Latches */
	PORT_SR &= ~(1 << PIN_RCK);	/* RCK auf 0 */
	PORT_SR |=  (1 << PIN_RCK);	/* RCK auf 1 */
	
    	/***********************************/
	/* naechste Zeile ermitteln/ausgeben */
    	/***********************************/
		/* Ebenen und Spalten weiterzaehlen */
	if (spalte < CUBE_Y - 1) {
		spalte++;
	} else {
		if (ebene  < CUBE_Z - 1) {ebene++;}  else {ebene = 0;}
		spalte = 0;
	}

		/* Zeile ins Schieberegister schieben */
	for (uint8_t i = 0; i < CUBE_X; i++) {
		if (led[(ebene*CUBE_Z*CUBE_Y)+(spalte*CUBE_Y)+i]) {
			PORT_SR |=  (1 << PIN_SER); /* 1 */
		} else {
			PORT_SR &= ~(1 << PIN_SER); /* 0 */
		}
		/* 0-1-Flanke an SCK schiebt Daten in den Registern um eine Stelle weiter */
		PORT_SR &= ~(1 << PIN_SCK); /* SCK auf 0 */
		PORT_SR |=  (1 << PIN_SCK); /* SCK auf 1 */
	}
	
		/* Ebene ins Schieberegister schieben */
	for (uint8_t i = 0; i < CUBE_Z; i++) {
		if (i != (CUBE_Z - ebene - 1)) {
			PORT_SR &= ~(1 << PIN_SER);	/* 0 */
		} else {
			PORT_SR |=  (1 << PIN_SER); /* 1 */		
		}
		/* 0-1-Flanke an SCK schiebt Daten in den Registern um eine Stelle weiter */
		PORT_SR &= ~(1 << PIN_SCK); /* SCK auf 0 */
		PORT_SR |=  (1 << PIN_SCK); /* SCK auf 1 */
	}
	
		/* Spalte ins Schieberegister schieben */
	for (uint8_t i = 0; i < CUBE_Y; i++) {
		if (i != (CUBE_Y - spalte - 1)) {
			PORT_SR &= ~(1 << PIN_SER); /* 0 */
		} else {
			PORT_SR |=  (1 << PIN_SER); /* 1 */		
		}	
		/* 0-1-Flanke an SCK schiebt Daten in den Registern um eine Stelle weiter */
		PORT_SR &= ~(1 << PIN_SCK); /* SCK auf 0 */
		PORT_SR |=  (1 << PIN_SCK); /* SCK auf 1 */
	}
	
		/* 0-1-Flanke an RCK bringt die Daten aus den Registern in die Latches */
	PORT_SR &= ~(1 << PIN_RCK);	/* RCK auf 0 */
	PORT_SR |=  (1 << PIN_RCK);	/* RCK auf 1 */
}


/* Kommando/Argumente extrahieren */
void SplitCmdArg(char * line, char ** cmd, char ** args)
{
    /* vorangestellte Leerzeichen ueberlesen */
    while (*line && *line <= ' ') {line++;}
    
    /* 1.Wort ist Kommando, in Grossbuchstaben konvertieren */
    *cmd = line;
    while (*line > ' ') {
        if (*line >= (unsigned char) 'a' && *line <= (unsigned char) 'z') { *line -= (unsigned char) 'a' - 'A';}
        line++;
    }
    
    /* Endemarkierung des Kommandowortes */
    if (*line) {*line++ = '\0';}
    
    /* Leerzeichen ueberlesen */
    while (*line && *line <= ' ') {++line;}
    
    /* dann kommen die Argumente */
    *args = line;
    while (*line && *line != '\r' && *line != '\n') {line++;}
    
    /* Endemarkierung der Argumente */
    *line = 0;
}


/* diverse Dinge initialisieren */
void cube_init(void)
{
	/* IP-Port */
	uip_listen(HTONS(CUBE_PORT));
	
	/* benoetigte Ports als Ausgaenge schalten */
	DDRA = (1 << DDA0) | (1 << DDA1) | (1 << DDA2) | (1 << DDA3) | (1 << DDA4) | (1 << DDA5);
	DDRC = (1 << DDC4) | (1 << DDC5) | (1 << DDC6) | (1 << DDC7);

	PORTC |= (1 << PC7);  /* SCL auf 1 */	

	/* Timer und Interrupt initialisieren*/
	TCNT1   =  timer_preload;	/* Preload-Wert     */
	TCCR1B |=  0x05;       		/* Vorteiler 1:1024 */
	TIMSK1 |= (1 << TOIE1);		/* Interrupt Timer1 */
	sei();
}



/*Steuerung via Kommandos via Netzwerk */
void cube_main(void)
{

	struct cube_cmd_state *s = &(uip_conn->appstate);

	char* cmd;
	char* arg;
	uint16_t arg_len;


	if(uip_rexmit())
		uip_send("cannot reximt. sorry.\n", 22);

	if(uip_connected()) {
//		s->textptr = build_date;
//		s->textlen = strlen(build_date);
	}

	if(uip_newdata()) {
		
		/* empfangenes Kommando auseinandernehmen */
		SplitCmdArg(uip_appdata, &cmd, &arg);
		
		/* ein neues Kommando beginnt? */
		if (! strcmp(cmd, CMD_SET_LED)      ||
		    ! strcmp(cmd, CMD_GET_XYZ)      ||
		    ! strcmp(cmd, CMD_VERSION)      ||
		    ! strcmp(cmd, CMD_HELP)         ||
		    ! strcmp(cmd, CMD_SET_PRELOAD)   ) 
		{
			s->state = NEW_CMD;
			s->arg_pos = 0;
			s->textlen = 0;
		}
		
		/* Kommando send_led noch nicht vollstaendig? */
		if (s->state == CMD_SET_LED_RECEIVE) {
			/* cmd hat hier die Nutzdaten! */
			arg_len = strlen(cmd); 
			for (uint16_t i = 0; i < arg_len; i++) {
				if ((s->arg_pos + i) < CUBE_XYZ) {
					led[s->arg_pos + i] = cmd[i]-'0';
				}
			}
			s->arg_pos = s->arg_pos + arg_len; 
			/* fehlt immer noch was? */
			if (s->arg_pos < CUBE_XYZ) {
				s->state = CMD_SET_LED_RECEIVE;
			} else {
				s->textptr = out_update_ok;
				s->textlen = strlen(out_update_ok);
			}
		}

		/* Schaltzustaende der LEDs uebernehmen */
    		if (! strcmp(cmd, CMD_SET_LED)) {
			arg_len = strlen(arg); 
			for (uint16_t i = 0; i < arg_len; i++) {
				if (i < CUBE_XYZ) {
					led[i] = arg[i]-'0';
				}
			}
			s->arg_pos = arg_len;
			/* fehlt noch was? */
			if (s->arg_pos < CUBE_XYZ) {
				s->state = CMD_SET_LED_RECEIVE;
			} else {
				s->textptr = out_update_ok;
				s->textlen = strlen(out_update_ok);
			}
		}
		
    		/* Dimension des LED-Cube */
		if(! strcmp(cmd, CMD_GET_XYZ)) {
			s->textptr = out_set_xyz;
			s->textlen = strlen(out_set_xyz);
		}

    		/* Schaltzustaende der LEDs abfragen */
	
    		/* Versionsabfrage */
		if(! strcmp(cmd, CMD_VERSION)) {
			s->textptr = build_date;
			s->textlen = strlen(build_date);
		}

    		/* zum TEST ... Preload setzen */
		if(! strcmp(cmd, CMD_SET_PRELOAD)) {
			timer_preload = atoi(arg);
		}

    		/* Hilfe */
		if(! strcmp(cmd, CMD_HELP)) {
		// hier koennte man einen Hilfetext ausgeben...
		}

	} 

	
	/* gibt es etwas zurueckzusenden? */
	if (uip_newdata()   ||
	    uip_rexmit()    ||
	    uip_acked()     ||
	    uip_poll()      ||
	    uip_connected())  {
		if (s->textlen > 0) {
			uip_send(s->textptr, s->textlen);
			s->textlen = 0;
		}
	}
}
